# -*- coding: utf-8 -*-
from PyQt5.QtCore import QUrl, Qt, QLocale
from PyQt5.QtGui import QIcon, QDesktopServices
from PyQt5.QtWidgets import QAction, QMessageBox, QInputDialog, \
    QDialog, QFileDialog, QApplication
from qgis.core import QgsVectorLayer, QgsProject, QgsGeometry, QgsFeature
from qgis.gui import QgsMapToolEmitPoint, QgsAttributeDialog
from qgis.core import QgsPoint, QgsFeatureRequest
from .db import get_connection
from .resources import *
from .paggf_layers import LoadedLayers
from .paggf_utils import PAGGFUtils, get_background_layer
import os.path
import tempfile
import inspect
from .layout.LayoutValidator import LayoutValidator
from collections import OrderedDict
from .ui.CID_munip_list import CIDMunipListDlg
from .ui.plugin_properties import PluginPropertiesDialog
from .ui.ALL_geom_dockwidget import ALLGeomDW
from .ui.select_municipality import SelectMunicipalityDlg
from .search_by_attribute import SearchByAttributeDialog
from .ui.download_sketch_dlg import DownloadSketch
from .ui.CID_hill_sketch import hill_sketch_dlg
from . import paggf_version
from .old_version_compatibility import geom_as_wkt, geom_from_coord_xy, \
    geom_from_point_xy

from .load_parcel_wizard import LoadIDCPolygonWizard
from .db_store_layout_items import Loader as StoreLayoutLoader

# import custom functions to use in QGIS expressions
from .expression_functions import *


class PAGGFToolbar(object):
    toolbars = dict()
    plugin = None

    @classmethod
    def create(cls, name, ui_name):
        """

        Keyword arguments:
        name -- Id of the toolbar
        ui_name -- Name displayed in the UI
        """
        PAGGFToolbar.toolbars[name] = cls.plugin.iface.addToolBar(ui_name)
        PAGGFToolbar.toolbars[name].setObjectName(ui_name)

    @classmethod
    def create_from_list(cls, l):
        for name, ui_name in l:
            cls.create(name, ui_name)

    @classmethod
    def delete(cls, n):
        del cls.toolbars[n]


class PAGGFAction(object):
    plugin = None  # Acces au plugin
    active = None  # si une action reste active a la fin du callback
    # la désactiver au lancement de la nouvelle action
    actions = dict()  # Lancement des actions

    def __init__(
            self, name, config, icon_path=None, callback=None,
            before_callback=None, after_callback=None,
            active_after_callback=False, stop_action=None
    ):
        self.name = name
        self.config = config
        self.icon_path = icon_path
        self.callback = callback
        self.before_callback = before_callback  # to execute before callback
        self.after_callback = after_callback  # to execute after callback
        self.active_after_callback = active_after_callback
        self.stop_action = stop_action

        self.action = None  # au debut action pas activee
        PAGGFAction.actions[self.name] = self

    def toolbar_name(self):
        if self.config and 'tb_name' in self.config:
            return self.config['tb_name']
        return None

    def activate(self):
        icon = QIcon(self.icon_path)
        self.action = QAction(
            icon, self.config['menu_title'],
            PAGGFAction.plugin.iface.mainWindow())
        self.action.triggered.connect(self.exec_callback)

        tb_name = self.toolbar_name()
        if tb_name and tb_name in PAGGFToolbar.toolbars:
            PAGGFToolbar.toolbars[tb_name].addAction(self.action)

        self.action.setCheckable(True)

    def exec_callback(self, *args, **kwargs):
        prev_active_action = PAGGFAction.active

        if PAGGFAction.active:
            PAGGFAction.active.action.setChecked(False)

            if PAGGFAction.active.stop_action:
                PAGGFAction.active.stop_action()
                PAGGFAction.active = None

        if prev_active_action == self:  # On desactive l'action car second
            # click sur l'icone
            pass
        else:
            self.exec_before_callback()

            sig = inspect.signature(self.callback)

            if len(sig.parameters) == 0:
                self.callback()
            else:
                self.callback(*args, **kwargs)

            self.exec_after_callback()

            if self.active_after_callback:
                PAGGFAction.active = self

    def exec_before_callback(self):
        if self.before_callback:
            self.before_callback()

    def exec_after_callback(self):
        if self.after_callback:
            self.after_callback()


class ClickDlgAction(PAGGFAction):
    """
    Action that is trigger when clicking on a icon, then display a dlg,
    then finish.
    """
    def __init__(self, *args, **kwargs):
        super(ClickDlgAction, self).__init__(*args, **kwargs)

        self.after_callback = self.after_callback_fct

    def after_callback_fct(self):
        self.action.setChecked(False)


def dirty_layer(layer):
    """
    Modifiy a layer so that it is considered dirty by QGIS
    and the current transaction won't be empty
    """
    for feature in layer.getFeatures():
        # sets id to id just to dirty buffer
        layer.editBuffer().changeAttributeValue(
            feature.id(),
            feature.fieldNameIndex('id'),
            feature.attribute('id'))
        break

    """# ugly hack:
    # we make sure the snap index is invalidated
    # by changing the subsetstring
    ss = layer.subsetString()
    if ss == '':
        layer.setSubsetString("id is not null")
    else:
        layer.setSubsetString("")
    """


def mogrify_sql(sql, args):
    return get_connection().cursor().mogrify(sql, args).decode('utf8')


def execute_sql_on_layer(layer, sql):
    """
    Execute a SQL query on the database of the given layer.
    The query will be inserted in the transaction object of the
    layer
    """
    # switch to edit mode if not already the case
    layer.startEditing()
    transaction = layer.dataProvider().transaction()
    transaction.executeSql(sql, isDirty=True)


def save_layers_edits():
    """
    Checks whether a layer needs to be saved.
    Returns False if a cancellation is required
    """
    for layer in QgsProject.instance().mapLayers().values():
        if not isinstance(layer, QgsVectorLayer):
            continue
        if layer.editBuffer() is not None:
            r = QMessageBox.question(None, "Sauvegarder les changements ?",
                                     "Des éditions sont en cours. Sauvegarder les changements avant de continuer ?",
                                     QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
            if r == QMessageBox.Yes:
                layer.commitChanges()
            elif r == QMessageBox.No:
                layer.rollBack()
            else:
                return False
            break
    return True


def icon_path(file_name):
    return os.path.join(os.path.dirname(__file__), "icons", file_name)


class PAGGFCore(object):
    """Encodes the CID inventory"""

    def __init__(self, iface, version, menu_title):
        """
        Constructor

        Keyword arguments:
        iface -- the QgsInterface
        version -- version of the plugin (BM, CID or EF see paggf_version.py)
        menu_title -- the tile of the menu
        """
        self.iface = iface
        self.version = version
        self.menu_title = menu_title

        self.mapCanvas = iface.mapCanvas()

        self.load_parcel_dlg = None

        self.layers = LoadedLayers()

        self.maptool = None

        self.actions = dict()  # Dictionary where are stored the actions
        self.actions_configuration = OrderedDict()  # configuration of the
        # PAGGF actions

        self.edit_geom_action_config = OrderedDict()

        self.utils = PAGGFUtils(self.layers)

        self.action_after_parcel_selection = {
            'delete_parcel': (
                icon_path("delete_parcel.svg"),
                self.delete_parcel, {}),
            'edit_data': (
                icon_path("edit_data.svg"),
                self.edit_data, {}),
            'archive_parcel': (
                icon_path("archive_parcel.svg"),
                self.archive_parcel_at, {
                    'menu_title': 'Archiver une parcelle'
                }),
            'copy_parcel': (
                icon_path("copy_parcel.svg"),
                self.copy_parcel_at, {
                    'menu_title': 'Copier une parcelle archivée'
                }),
            'pv_upload_signatures': (
                icon_path("merge_doc.svg"),
                self.pv_upload_signatures, {}),
            'generate_and_validate_parcel_sketch': (
                icon_path("parcel_sketch.svg"),
                self.get_generate_and_validate_parcel_sketch(
                    'generate_and_validate_parcel_sketch'), {}),
            'download_parcel_sketch': (
                icon_path("sketch_pdf.svg"),
                self.download_parcel_sketch, {}),
            'pv_generate_and_print': (
                icon_path("print.svg"),
                self.pv_generate_and_print, {}),
            'generate_and_validate_parcel_sketch_no': (
                icon_path("house_sketch.svg"),
                self.get_generate_and_validate_parcel_sketch(
                    'generate_and_validate_parcel_sketch_no'), {}),
            'new_ef': (
                icon_path("load_attributes_from_file.svg"),
                self.new_ef, {}),

        }

        PAGGFAction.plugin = self
        PAGGFToolbar.plugin = self

    def enable_paggf_action(self, action_id, configuration):
        """
        Enable a paggf action. To be called during __init__.

        Keyword arguments:
        action_id -- The id of the action
        configuration -- its configuration
        """
        self.actions_configuration[action_id] = configuration

    def config_edit_geom_action(self, action_id, config):
        self.edit_geom_action_config[action_id] = config

    def paggf_menu_content(self):
        "To be defined by derived classes"
    def toolbar_list_def(self):
        "To be defined by derived classes"

    def create_paggf_menu(self):
        """
        Create the PAGGF menu (the tree of the menu is defined in the
        varialbe self.paggf_menu_content)
        """
        self.paggf_menu = self.iface.mainWindow().menuBar().addMenu(
            self.menu_title)
        self.paggf_menu.setObjectName('mPAGGF')

        def def_recursive_create_menu(parent, children):
            """
            Ajout recursif d'un menu
            parent  -> deja un menu
            la suite (à ajouter) soit
                - une liste (sous-menu)
                - une action (fin du la branche du menu)
                - un label (action pas encore ecrite)
            """
            for child in children:
                if isinstance(child, PAGGFAction):  # fin de l'arble ->
                    # ajoute l'action
                    parent.addAction(child.action)
                elif type(child) == tuple:  # ajout sous menu
                    c_label, c_children = child
                    c_menu = parent.addMenu(c_label)
                    def_recursive_create_menu(c_menu, c_children)
                elif type(child) == str:  # Ajout d'une action non ecrite
                    action = parent.addAction(child)
                    action.setEnabled(False)
                else:  # Error
                    print(child)
                    print(type(child))

        def_recursive_create_menu(self.paggf_menu, self.paggf_menu_content())

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        PAGGFToolbar.create_from_list(self.toolbar_list_def())

        for action_name in self.actions_configuration:
            config = self.actions_configuration[action_name]

            if action_name in self.action_after_parcel_selection:
                icon, callback, opt_conf = self.action_after_parcel_selection[
                    action_name]
                config_with_opt = {**config, **opt_conf}
                ClickDlgAction(
                    action_name, config_with_opt, icon,
                    self.check_selection_then_trigger_action_factory(
                        config['pg_layers'],
                        callback))

            elif action_name == 'load_parcel_from_file':
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("load_parcel_from_file.svg"),
                    self.get_load_parcel_callback(config))

            elif action_name == 'load_polygon_from_file':
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("load_polygon_from_file.svg"),
                    self.load_polygon_from_file)

            elif action_name == 'add_to_parcel_from_file':
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("add_to_parcel_from_file.svg"),
                    self.get_load_parcel_callback(config, merge_in_parcel=True)
                )

            elif action_name == 'load_parcel_manually':
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("load_parcel_manually.svg"),
                    self.get_load_parcel_callback(config))
            elif action_name == 'layout_map_munip_inventory':
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("munip_inventory.svg"),
                    self.layout_map_munip_inventory)

            elif action_name == 'layout_munip_list':
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("munip_list.svg"),
                    self.layout_munip_list)

            elif action_name == 'layout_hill_map':
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("hill_map.svg"),
                    self.layout_hill_map)

            elif action_name == "plugin_properties":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("properties.svg"),
                    self.display_properties_dlg)

            elif action_name == "edit_geom":
                PAGGFAction(
                    action_name,
                    config,
                    icon_path("edit_geom.svg"),
                    self.activate_desactivate_geom_edit,
                    stop_action=self.pan_trigger)

            elif action_name == "edit_internal_objects":
                PAGGFAction(
                    action_name,
                    config,
                    icon_path("edit_internal_objects.svg"),
                    self.activate_internal_objects,
                    stop_action=self.pan_trigger)

            elif action_name == "save_edits":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("save_edits.svg"),
                    self.save_edits,
                    stop_action=self.pan_trigger)

            elif action_name == "search":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("search_valid.svg"),
                    self.search_valid)

            elif action_name == "search_invalid":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("search.svg"),
                    self.search_invalid)

            elif action_name == "attribute_table":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("attributes.svg"),
                    self.attribute_table)

            elif action_name == "export_db":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("export_db.svg"),
                    self.export_db_to_dir)

            elif action_name == "import_db":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("import_db.svg"),
                    self.import_db_from_dir)

            elif action_name == "sync_db":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("sync_db.svg"),
                    self.sync_db)

            elif action_name == "edit_efpf_data":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("edit_data.svg"),
                    self.edit_efpf_data)

            elif action_name == "generate_efpf":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("print.svg"),
                    self.generate_efpf)

            elif action_name == "select":
                ClickDlgAction(
                    action_name,
                    config,
                    icon_path("select_rectangle.svg"),
                    self.on_select)

            else:
                print('action {} not found'.format(create_action))

            PAGGFAction.actions[action_name].activate()

        self.create_paggf_menu()
        self.init_geom_edit_dockwidget()

        dlg = self.welcome_dlg_class()
        dlg.exec_()

    def on_select(self):
        # select on every visible layers, except "ref" layers
        # exclude also the mask layer
        layers = [l for l in self.iface.mapCanvas().layers()
                  if isinstance(l, QgsVectorLayer) and "ref" not in l.source() and l.providerType() != "memory"]
        from .maptool_select import MapToolSelectRectangle
        maptool = MapToolSelectRectangle(self.iface.mapCanvas(), layers)
        self.iface.mapCanvas().setMapTool(maptool)

    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for paggf_action in PAGGFAction.actions.values():
            self.iface.removeToolBarIcon(paggf_action.action)
        # remove the toolbar

        toolbar_names = [n for n in PAGGFToolbar.toolbars]
        for n in toolbar_names:
            PAGGFToolbar.delete(n)

        # remove the menu
        self.paggf_menu.clear()
        self.paggf_menu.deleteLater()

    def get_load_parcel_callback(self, config, merge_in_parcel=False):
        def ret():
            if not merge_in_parcel:
                self.load_parcel_wizard(config)
            else:
                self.add_vertices_to_parcel(config)
        return ret

    def load_parcel_wizard(self, config):
        WizardClass = config['dlg_class']
        wiz = WizardClass(config["from_file"], merge_in_parcel=False)
        r = wiz.exec_()
        if r == QDialog.Rejected:
            return

        if not save_layers_edits():
            return

        conn = get_connection()
        cur = conn.cursor()

        id_parcel = wiz.id_parcel()
        if id_parcel is None:
            # add a new parcel
            cur.execute("\
            insert into gis.parcels_to_valid (id, b_id, id_sub_hill_avenue, num_parcel) \
            select gid, gid, %s, %s \
            from (select gen_random_uuid() as gid) t \
            returning id",
                        (wiz.subhill(), wiz.parcel_number()))
            id_parcel = cur.fetchone()[0]

        # populate gis.vertices_to_import
        cur.execute("delete from gis.vertices_to_import")
        values = []
        for f, geom in wiz.get_features():
            wkt = 'srid=32735;' + geom_as_wkt(geom)
            name = f[wiz.name_field()]
            values.append(name)
            values.append(wkt)

        query = "insert into gis.vertices_to_import (name, geom) values "
        query += ','.join(['(%s, st_force2d(%s))'] * (len(values)//2))
        cur.execute(query, values)

        # call the import function
        cur.execute("select gis.{}('{}')".format(
            config["import_function_name"],
            id_parcel))

        conn.commit()

        for l in config['layers_to_update']:
            lyr = self.layers.get('gis', l)
            # setsubstring will call updateFeatureCount
            # this way, the extent is correctly updated
            lyr.setSubsetString("")
            lyr.updateExtents()

        # get featureid from uuid
        l = self.layers.get('gis', config['layer_to_zoom'])
        for f in l.getFeatures(
            QgsFeatureRequest().setFilterExpression(
                "id = '{}'".format(id_parcel))):
            pass

        self.iface.mapCanvas().zoomToFeatureIds(l, [f.id()])

    def load_polygon_from_file(self):
        config = self.actions_configuration['load_polygon_from_file']
        wiz = LoadIDCPolygonWizard(self.iface.mainWindow())
        r = wiz.exec_()
        if r == QDialog.Rejected:
            return

        layer_out = self.layers.get("import", config["import_table"])
        layer_out.startEditing()
        ids = wiz.ids_parcel()
        if ids == []:
            return
        # (parcel_code, QgsFeatureId) in ids
        parcel_ids = [x[0] for x in ids]
        feats = []
        bbox = None
        for f in wiz.layer().getFeatures():
            if f['id'] in parcel_ids:
                fo = QgsFeature(layer_out.fields())
                fo.setGeometry(QgsGeometry(f.geometry()))
                if bbox is None:
                    bbox = f.geometry().boundingBox()
                else:
                    bbox.combineExtentWith(f.geometry().boundingBox())
                feats.append(fo)
        layer_out.addFeatures(feats)
        layer_out.deleteFeatures([x[1] for x in ids])

        # zoom on the new parcels
        if bbox is not None:
            self.iface.mapCanvas().zoomToFeatureExtent(bbox)

    def add_vertices_to_parcel(self, config):
        WizardClass = config['dlg_class']
        wiz = WizardClass(from_file=True, merge_in_parcel=True)
        r = wiz.exec_()
        if r == QDialog.Rejected:
            return

        id_parcel = wiz.id_parcel()
        parcel_layer = self.layers.get('gis', config['parcel_layer'])
        for f in parcel_layer\
            .getFeatures(QgsFeatureRequest()\
                         .setFilterExpression("id = '{}'".format(id_parcel))):
            num_parcel = f['num_parcel']
        layer = self.layers.get('gis', config['vertex_layer'])
        expr = "a_parcel_id = '{}' and parcel_num_at_creation = {}".format(
            id_parcel, num_parcel)
        vertex_num = max([int(f['vertex_num_at_creation']) \
                          for f in layer.getFeatures(QgsFeatureRequest().setFilterExpression(expr))]) + 1

        layer.startEditing()
        feats = []
        for f, geom in wiz.get_features():
            fo = QgsFeature(layer.fields())
            fo['a_parcel_id'] = id_parcel
            fo['parcel_num_at_creation'] = num_parcel
            fo['vertex_num_at_creation'] = vertex_num
            fo['is_main'] = f[wiz.name_field()][0] == 'P'
            fo.setGeometry(QgsGeometry(f.geometry()))
            vertex_num += 1
            feats.append(fo)
        layer.addFeatures(feats)

        for l in config['layers_to_update']:
            lyr = self.layers.get('gis', l)
            lyr.updateExtents()

        # get featureid from uuid
        for f in parcel_layer.getFeatures(
            QgsFeatureRequest().setFilterExpression(
                "id = '{}'".format(id_parcel))):
            pass

        self.iface.mapCanvas().zoomToFeatureIds(
            parcel_layer,
            [f.id()])

    def select_point_then_trigger_action_factory(
            self, db_schema_name, db_table_name, action_to_trigger):

        layer = self.layers.get(db_schema_name, db_table_name)

        def on_feature_clicked(point_clicked):
            pt = geom_from_point_xy(point_clicked)
            pt_buffer = pt.buffer(
                (self.mapCanvas.mapUnitsPerPixel() * 3), 0)
            pt_rect = pt_buffer.boundingBox()

            req = QgsFeatureRequest(pt_rect).setFlags(
                QgsFeatureRequest.ExactIntersect)
            features = layer.getFeatures(req)

            feature = QgsFeature()
            if features.nextFeature(feature):
                # make sure the iterator is closed to avoid locks
                features.close()
                # call callback
                action_to_trigger(layer, point_clicked, feature)

        def select_point_then_trigger_action():
            vertex_layer = self.layers.get(db_schema_name, db_table_name)
            self.iface.setActiveLayer(vertex_layer)
            self.maptool = QgsMapToolEmitPoint(self.mapCanvas)
            self.maptool.canvasClicked.connect(on_feature_clicked)
            self.mapCanvas.setMapTool(self.maptool)

        return select_point_then_trigger_action

    def check_selection_then_trigger_action_factory(
            self, db_layers, action_to_trigger):
        layers = [
            self.layers.get(db_schema_name, db_table_name)
            for db_schema_name, db_table_name in db_layers]

        def check_selection_then_trigger_action():
            triggered_action = False

            for layer in layers:
                if layer.selectedFeatureCount() == 1:
                    feature = layer.selectedFeatures()[0]
                    action_to_trigger(layer, QgsPoint(), feature)
                    triggered_action = True
                    break

            if not triggered_action:
                # si feature pas trouve
                self.iface.messageBar().pushCritical(
                    'Erreur',
                    'Veuillez sélectionner une entité sur la couche "{}" pour \
cette action.'.format(layers[0].name()))
                self.iface.setActiveLayer(layers[0])

        return check_selection_then_trigger_action

    def add_vertex_at(self, layer, point_clicked, edge):
        config = self.edit_geom_action_config['add_vertex']
        add_vertex_layer = self.layers.get(
            config['pg_schema_name'],
            config['pg_table_name'])
        add_vertex_layer.startEditing()

        pt_wkt_txt = "'srid=32735;point({} {})'::geometry"\
            .format(point_clicked.x(), point_clicked.y())
        sql = "SELECT gis.add_vertex_to_polygon('{}'::UUID,'{}'::UUID,'{}'::UUID,'{}'::UUID,{})"\
            .format(
                edge['id_polygon'],
                edge['id_parcel'], edge['id_vertex_before'],
                edge['id_vertex_after'], pt_wkt_txt)
        execute_sql_on_layer(add_vertex_layer, sql)
        add_vertex_layer.triggerRepaint()

    def add_isolated_vertex_at(self, layer, point_clicked, edge):
        config = self.edit_geom_action_config['add_isolated_vertex']
        add_vertex_layer = self.layers.get(
            config['pg_schema_name'],
            config['pg_table_name'])
        self.iface.setActiveLayer(add_vertex_layer)
        add_vertex_layer.startEditing()
        self.iface.actionAddFeature().trigger()

    def add_digitized_vertex_at(self, layer, point_clicked, edge):
        config = self.edit_geom_action_config['add_isolated_vertex']
        add_vertex_layer = self.layers.get(
            config['pg_schema_name'],
            config['pg_table_name'])
        parcel_layer = self.layers.get(
            config['pg_schema_name'],
            config['parcel_layer'])
        self.iface.setActiveLayer(add_vertex_layer)
        add_vertex_layer.startEditing()
        x, ok = QInputDialog.getDouble(None, "Coordonnée X du sommet", "X")
        if not ok:
            return
        y, ok = QInputDialog.getDouble(None, "Coordonnée Y du sommet", "Y")
        if not ok:
            return
        f = QgsFeature(add_vertex_layer.fields())
        f.setGeometry(geom_from_coord_xy(x, y))
        f["a_parcel_id"] = parcel_layer.selectedFeatures()[0]["id"]
        ad = QgsAttributeDialog(add_vertex_layer, f, False)
        if ad.exec_() == QDialog.Rejected:
            return
        f = ad.feature()
        add_vertex_layer.addFeatures([f])

    def move_vertex(self):
        config = self.edit_geom_action_config['move_vertex']
        vertex_layer = self.layers.get(config['pg_schema_name'],
                                       config['pg_table_name'])
        self.iface.setActiveLayer(vertex_layer)
        vertex_layer.startEditing()
        self.iface.actionMoveFeature().trigger()

    def delete_vertex_at(self, vertex_layer, point_clicked, feat):
        vertex_layer.startEditing()
        vertex_layer.deleteFeature(feat.id())
        self.mapCanvas.refresh()

    def edit_vertex_at(self, vertex_layer, point_clicked, feat):
        vertex_layer.startEditing()
        self.iface.openFeatureForm(vertex_layer, feat)
        self.mapCanvas.refresh()

    def add_ring(self):
        """
        Trigger the addRing action on the layer idc_parcels_to_valid. This
        action allow the user to create ring into a polygon
        """
        config = self.edit_geom_action_config['add_ring']
        parcels_layer = self.layers.get(config['pg_schema_name'],
                                        config['pg_table_name'])
        self.iface.setActiveLayer(parcels_layer)
        parcels_layer.startEditing()
        self.iface.actionAddRing().trigger()

    def delete_ring(self):
        """
        Trigger the deleteRing action on the layer idc_parcels_to_valid. This
        action allow the user to delete a ring into a polygon
        """
        config = self.edit_geom_action_config['delete_ring']
        parcels_layer = self.layers.get(config['pg_schema_name'],
                                        config['pg_table_name'])
        self.iface.setActiveLayer(parcels_layer)
        parcels_layer.startEditing()
        self.iface.actionDeleteRing().trigger()

    def reshape(self):
        """
        Trigger the reshape / the cut feature action on the layer
        idc_parcels_to_valid. This action allow the user to cut external part
        of a polygone
        """
        config = self.edit_geom_action_config['reshape']
        parcels_layer = self.layers.get(config['pg_schema_name'],
                                        config['pg_table_name'])
        self.iface.setActiveLayer(parcels_layer)
        parcels_layer.startEditing()
        reshapeAction = [
            o for o in self.iface.mainWindow().findChildren(QAction)
            if o.objectName() == 'mActionReshapeFeatures'][0]
        reshapeAction.trigger()

    def edit_data(self, parcels_layer, point_clicked, feature):
        self.iface.setActiveLayer(parcels_layer)
        parcels_layer.startEditing()
        self.iface.openFeatureForm(parcels_layer, feature)

    def delete_parcel(self, parcel_layer, point_clicked, feature):
        parcel_layer.startEditing()
        parcel_layer.deleteFeature(feature.id())
        self.iface.mapCanvas().refreshAllLayers()

    def archive_parcel_at(self, layer, point_clicked, parcel):
        config = self.actions_configuration["archive_parcel"]
        pg_schema_name = None
        pg_table_name = None

        for schema_l, name_l in config['pg_layers']:
            l = self.layers.get(schema_l, name_l)
            if layer == l:
                pg_schema_name = schema_l
                pg_table_name = name_l

        r = QMessageBox.warning(None,
                                "Confirmation d'archivage",
                                "L'archivage d'une parcelle est une opération \
irréversible. Confirmez-vous l'archivage ?",
                                QMessageBox.Yes | QMessageBox.No)
        if r == QMessageBox.No:
            return

        event = QInputDialog.getText(
            None, "Motif d'archivage", "Motif d'archivage")

        sql = mogrify_sql(("update {}.{} set " +
                           "validity_period = tstzrange(now(), 'infinity', '[)'), " +
                           "validity_event = %s where id = %s::UUID").format(
                               pg_schema_name,
                               pg_table_name),
                          (event, parcel['id']))

        execute_sql_on_layer(layer, sql)

        self.iface.mapCanvas().refreshAllLayers()

    def copy_parcel_at(self, parcel_layer, point_clicked, parcel):
        execute_sql_on_layer(parcel_layer,
                             mogrify_sql("select copy_object('land.parcels', %s::UUID)",
                                         (parcel['id'],)))

        self.iface.mapCanvas().refreshAllLayers()

    def pv_upload_signatures(self, parcel_layer, point_clicked, feature):
        from .ui.ALL_merge_pv_and_signatures import CIDReportMergeDlg, EFReportMergeDlg
        if self.version == paggf_version.CID:
            self.dlg = CIDReportMergeDlg(feature, parcel_layer)
        else:
            self.dlg = EFReportMergeDlg(feature, parcel_layer)
        self.dlg.exec_()

    def pv_generate_and_print(self, layer, point_clicked, parcel):
        config = self.actions_configuration['pv_generate_and_print']
        self.dlg = config['dlg_class'](self.iface, layer, parcel)

        if 'scale_step' in config:
            self.dlg.map_scale_step = config['scale_step']

        if 'zoom_to_feature_scale' in config:
            self.dlg.zoom_to_feature_scale = config['zoom_to_feature_scale']

        self.dlg.exec_()

    def pan_trigger(self):
        self.iface.actionPan().trigger()

    def get_generate_and_validate_parcel_sketch(self, action_name):
        def slot(layer, point_clicked, feature):
            config = self.actions_configuration[action_name]

            if self.version == paggf_version.BM:
                output_filename = '{}_{}_CROQUIS_BM'.format(
                    feature['idc_parcel_code'], feature['hill_district_name'])
                template_file = 'layout/sketch_bm_A4_vlayout.qpt'
            elif self.version == paggf_version.CID:
                output_filename = '{}_{}_CROQUIS_CID'.format(
                    feature['idc_parcel_code'], feature['hill_district_name'])
                template_file = 'layout/sketch_cid_A4_vlayout.qpt'
            else:
                output_filename = '{}_{}_CROQUIS_EFTF'.format(
                    feature['idc_parcel_code'], feature['hill_district_name'])
                template_file = 'layout/sketch_other_A4_vlayout.qpt'

            template_path = os.path.join(
                self.utils.plugin_dir, template_file)

            self.cv = LayoutValidator(
                feature, template_path, self.iface, zoom_to_feature=True,
                scale_step=250, output_png=True, output_pdf=True,
                output_filename=output_filename, load_feature_in_sub_map=True)

            self.cv.version = self.version

            if 'export_after_generation' in config and config['export_after_generation']:
                # affichage du PV en pdf
                pass
            else:
                self.cv.callback = self.cv.get_callback_to_store_into_feature(
                    layer, feature, 'doc_pdf_parcel_sketch',
                    'doc_png_parcel_sketch')

            self.cv.add_substitutions({
                'NORTH_ARROW_PATH': os.path.join(
                    self.utils.plugin_dir, 'img/north.png'),
            })

            if 'commission_reference' in feature:
                self.cv.add_substitution(
                    'cid_number', feature['commission_reference'])

            if self.version == paggf_version.CID:
                layers_config_carto = []

                for layer, style, style_substitution_map in [
                    (
                        self.layers.get('gis', 'idc_parcels_to_valid'),
                        os.path.join(
                            self.utils.plugin_dir, "layer_style/idc_only_current_parcel_sketch_A4.qml"),
                        {'current_parcel_id': lambda f: f['id']}),
                    (
                        self.layers.get('gis', 'current_parcels'),
                        os.path.join(
                            self.utils.plugin_dir, "layer_style/idc_only_current_parcel_sketch_A4.qml"),
                        {'current_parcel_id': lambda f: f['id']}),
                    (
                        self.layers.get('gis', 'idc_vertices_to_valid'),
                        os.path.join(
                            self.utils.plugin_dir, "layer_style/idc_only_current_vertex_sketch_A4.qml"),
                        {'current_parcel_id': lambda f: f['id']}),
                    (
                        self.layers.get('gis', 'current_idc_vertices'),
                        os.path.join(
                            self.utils.plugin_dir, "layer_style/idc_only_current_vertex_sketch_A4.qml"),
                        {'current_parcel_id': lambda f: f['id']}),
                ]:
                    if layer is not None:
                        layers_config_carto.append((layer, style, style_substitution_map))


                if 'with_ortho' in config and config['with_ortho']:
                    layers_config_carto.append((get_background_layer(), None))

                self.cv.layers_config = {
                    'carto': layers_config_carto
                }

            elif self.version == paggf_version.BM:
                fr_locale = QLocale(QLocale.French, QLocale.France)
                if feature['end_date']:
                    delivery_date_day_month = fr_locale.toString(
                        feature['end_date'], 'MMMM yyyy')
                else:
                    delivery_date_day_month = 'NULL'

                self.cv.add_substitution(
                    'place_and_date',
                    "{}, {}".format(
                        feature['municipality_name'], delivery_date_day_month))
                self.cv.add_substitution(
                    'calculated_area_m2_for_sketch',
                    "{:.2f}".format(feature['calculated_area_m2']))

                self.cv.layers_config = {
                    'carto': [
                        (
                            self.layers.get('gis', 'parcels_to_valid'),
                            os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_only_current_parcel_A4.qml"),
                            {'current_parcel_idc_code': lambda f: f['idc_parcel_code']}),
                        (
                            self.layers.get('gis', 'vertices_to_valid'),
                            os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_only_current_vertex_A4.qml"),
                            {'current_parcel_id': lambda f: f['id']}),
                        (self.layers.get(
                            'gis', 'parcel_buildings_to_valid'), os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_building_A4.qml")),
                        (self.layers.get(
                            'gis', 'parcel_rivers_to_valid'), os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_river_A4.qml")),
                        (self.layers.get(
                            'gis', 'parcel_roads_to_valid'), os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_road_A4.qml")),
                        (
                            self.layers.get('gis', 'current_parcels'),
                            os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_only_current_parcel_A4.qml"),
                            {'current_parcel_idc_code': lambda f: f['idc_parcel_code']}),
                        (
                            self.layers.get('gis', 'current_vertices'),
                            os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_only_current_vertex_A4.qml"),
                            {'current_parcel_id': lambda f: f['id']}),
                        (self.layers.get(
                            'gis', 'current_parcel_buildings'), os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_building_A4.qml")),
                        (self.layers.get(
                            'gis', 'current_parcel_rivers'), os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_river_A4.qml")),
                        (self.layers.get(
                            'gis', 'current_parcel_roads'), os.path.join(
                                self.utils.plugin_dir,
                                "layer_style/bm_road_A4.qml")),
                    ]
                }

            if 'scale_step' in config:
                self.cv.scale_step = config['scale_step']

            if 'zoom_to_feature_scale' in config:
                self.cv.zoom_to_feature_scale = config['zoom_to_feature_scale']

            self.cv.exec_()

            # for "Préparer et exporter/imprimer un extrait de\croquis de parcelle ID"
            if 'do_not_load_from_db' in config and config['do_not_load_from_db']:
                pass
            else:
                db_layout_storage = StoreLayoutLoader.get(self.version)
                db_layout_storage.load(feature, self.cv.layout)

        return slot

    def download_parcel_sketch(self, layer, point_clicked, feature):
        self.dlg = DownloadSketch(layer, feature)
        self.dlg.exec_()

    def layout_map_munip_inventory(self):
        def callback(output_pdf_path, output_png_path):
            self.utils.current_province_layer.clear()
            self.utils.mask_layer.create_or_update()
            QDesktopServices.openUrl(
                QUrl.fromLocalFile(output_png_path))
            self.mapCanvas.refresh()

        municipality = self.utils.get_current_municipality()
        province = self.utils.get_current_province()

        if municipality is None:
            dlg = SelectMunicipalityDlg()
            if dlg.exec_() == QDialog.Rejected:
                return
            municipality = self.utils.get_municipality(dlg.municipality_id())
            province = self.utils.get_province_by_munip_id(
                dlg.municipality_id())

        template_path = os.path.join(
            self.utils.plugin_dir,
            'layout/CID_A0_municipality_inventory_vlayout.qpt')

        self.cv = LayoutValidator(
            municipality, template_path, self.iface, zoom_to_feature=True,
            output_png=True, callback=callback)

        self.utils.mask_layer.clear()
        self.utils.mask_layer.create_or_update(municipality['id'])
        mask_layer = self.utils.mask_layer.get()
        current_province_layer = self.utils.current_province_layer.create_or_update()

        self.cv.layers_config = {
            'carto': [
                (mask_layer, os.path.join(
                    self.utils.plugin_dir, "layer_style/mask_white.qml")),
                (self.layers.get('ref', 'hill_districts'), os.path.join(
                    self.utils.plugin_dir,
                    "layer_style/munip_map_hill_districts.qml")),
                (self.layers.get('gis', 'encoded_idc_parcels'), os.path.join(
                    self.utils.plugin_dir,
                    "layer_style/munip_map_encoded_idc_parcels.qml")),
                (self.layers.get('ref', 'provinces'), None),
                (self.layers.get('ref', 'municipalities'), None),
                (get_background_layer(), None)],
            'Country Map': [
                (mask_layer, os.path.join(
                    self.utils.plugin_dir,
                    "layer_style/map_country_current_munip.qml")),
                (self.layers.get('ref', 'provinces'), os.path.join(
                    self.utils.plugin_dir,
                    "layer_style/map_country_provinces.qml")),
                (current_province_layer, os.path.join(
                    self.utils.plugin_dir,
                    "layer_style/map_country_current_province.qml")),
                (self.layers.get('ref', 'municipalities'), os.path.join(
                    self.utils.plugin_dir,
                    "layer_style/map_country_municipalities.qml")),
            ]
        }

        self.cv.substitution_map['municipality'] = municipality['name']
        self.cv.substitution_map['province'] = province['name']

        self.cv.substitution_map['LOGO_GIZ_PATH'] = os.path.join(
            self.utils.plugin_dir, 'img/logo_giz.png')
        self.cv.substitution_map['LOGO_EUROPE_PATH'] = os.path.join(
            self.utils.plugin_dir, 'img/logo_europe.png')
        self.cv.substitution_map['LOGO_PAGGF_PATH'] = os.path.join(
            self.utils.plugin_dir, 'img/logo_paggf.png')
        self.cv.substitution_map['LOGO_BURUNDI_PATH'] = os.path.join(
            self.utils.plugin_dir, 'img/Blason_du_Burundi.svg')
        self.cv.substitution_map['NORTH_ARROW_PATH'] = os.path.join(
            self.utils.plugin_dir, 'img/north.png')

        self.cv.exec_()

    def layout_munip_list(self):
        municipality_id = self.utils.get_current_municipality_id()

        if municipality_id is None:
            dlg = SelectMunicipalityDlg()
            if dlg.exec_() == QDialog.Rejected:
                return
            municipality_id = dlg.municipality_id()

        dlg = CIDMunipListDlg(municipality_id)
        dlg.exec_()

    def layout_hill_map(self):
        self.dlg = hill_sketch_dlg(self.iface)
        if self.dlg:
            self.dlg.exec_()

    def display_properties_dlg(self):
        dlg = PluginPropertiesDialog()
        if dlg.exec_() == QDialog.Accepted:
            # update municipality mask
            self.utils.mask_layer.create_or_update()
            self.iface.mapCanvas().refreshAllLayers()

    def activate_desactivate_geom_edit(self):
        self.geom_edit_dockwidget.activate_desactivate()

    def activate_internal_objects(self):
        lyr = self.layers.get("gis", "parcel_roads_to_valid")
        self.iface.setActiveLayer(lyr)
        lyr.startEditing()

    def save_edits(self):
        save_layers_edits()

    def init_geom_edit_dockwidget(self):
        main_window = self.iface.mainWindow()
        self.geom_edit_dockwidget = ALLGeomDW(self.pan_trigger, main_window)
        main_window.addDockWidget(
            Qt.LeftDockWidgetArea, self.geom_edit_dockwidget)

        for action_name in self.edit_geom_action_config:
            config = self.edit_geom_action_config[action_name]

            icon_ext = 'svg'
            if action_name == 'add_vertex':
                icon_ext = 'png'
                title = 'Ajouter un sommet sur un segment'
                callback = self.select_point_then_trigger_action_factory(
                    config['pg_schema_name'],
                    config['segments_view'],
                    self.add_vertex_at)

            if action_name == 'add_isolated_vertex':
                title = 'Ajouter un sommet isolé'
                callback = self.check_selection_then_trigger_action_factory(
                    [(config["pg_schema_name"], config["parcel_layer"])],
                    self.add_isolated_vertex_at)

            if action_name == 'add_digitized_vertex':
                title = 'Numériser un nouveau sommet'
                callback = self.check_selection_then_trigger_action_factory(
                    [(config["pg_schema_name"], config["parcel_layer"])],
                    self.add_digitized_vertex_at)

            elif action_name == 'move_vertex':
                title = 'Déplacer les sommets'
                icon_ext = 'png'
                callback = self.move_vertex

            elif action_name == 'delete_vertex':
                title = 'Supprimer les sommets'
                icon_ext = 'png'
                callback = self.select_point_then_trigger_action_factory(
                    config['pg_schema_name'],
                    config['pg_table_name'], self.delete_vertex_at)

            elif action_name == 'edit_vertex':
                title = "Editer les attributs d'un sommet"
                icon_ext = 'png'
                callback = self.select_point_then_trigger_action_factory(
                    config['pg_schema_name'],
                    config['pg_table_name'], self.edit_vertex_at)

            elif action_name == 'add_ring':
                title = 'Ajouter un anneau'
                callback = self.add_ring

            elif action_name == 'delete_ring':
                title = 'Supprimer  un anneau'
                callback = self.delete_ring

            elif action_name == 'reshape':
                title = 'Remodeler les entités'
                callback = self.reshape

            elif action_name == 'undo':
                title = 'Annuler'
                callback = lambda: self.iface\
                                       .mainWindow()\
                                       .findChild(QAction, "mActionUndo")\
                                       .trigger()

            elif action_name == 'redo':
                title = 'Refaire'
                callback = lambda: self.iface\
                                       .mainWindow()\
                                       .findChild(QAction, "mActionRedo")\
                                       .trigger()

            icon_path = os.path.join(
                os.path.dirname(__file__), "icons", "geom_edit",
                'icon_{}.{}'.format(action_name, icon_ext))

            self.geom_edit_dockwidget.add_action(
                action_name, icon_path, callback, title)

        self.geom_edit_dockwidget.hide()

    def search_valid(self):
        config = self.actions_configuration["search"]
        lyr = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])
        print("lyr", lyr.name())
        self.search_dlg = SearchByAttributeDialog(
            lyr, self.iface.mapCanvas(), config['hidden_fields'])
        self.search_dlg.show()

    def search_invalid(self):
        config = self.actions_configuration["search_invalid"]
        lyr = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])
        self.search_dlg = SearchByAttributeDialog(
            lyr, self.iface.mapCanvas(), config['hidden_fields'])
        self.search_dlg.show()

    def search_invalid(self):
        config = self.actions_configuration["search_invalid"]
        lyr = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])
        self.search_dlg = SearchByAttributeDialog(
            lyr, self.iface.mapCanvas(), config['hidden_fields'])
        self.search_dlg.show()

    def attribute_table(self):
        config = self.actions_configuration["attribute_table"]
        lyr = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])
        self.search_dlg = SearchByAttributeDialog(
            lyr, self.iface.mapCanvas(), config['hidden_fields'], display_whole_table = True)
        self.search_dlg.show()

    def export_db_to_dir(self):
        if not save_layers_edits():
            return
        from .sync import export_data_files

        dirname = QFileDialog.getExistingDirectory(
            None,
            "Répertoire d'export",
            "Choisissez le répertoire où exporter le contenu de la base locale"
        )
        if dirname is None:
            return

        conn = get_connection()
        cur = conn.cursor()
        export_data_files(cur, dirname)

        QMessageBox.information(
            None, "Export effectué", "Export effectué dans {}".format(dirname))

    def import_db_from_dir(self):
        if not save_layers_edits():
            return
        from .sync import import_data_files

        dirname = QFileDialog.getExistingDirectory(
            None,
            "Répertoire d'import",
            "Choisissez le répertoire depuis lequel importer le contenu de la base locale"
        )
        if dirname is None:
            return
        if QMessageBox.question(None, "Confirmation",
                                "L'import détruira toutes les données actuelles de la base locale pour " + \
                                "les remplacer par celles du répertoire sélectionné. Êtes-vous sûr de " + \
                                " vouloir continuer ?") == QMessageBox.No:
            return

        conn = get_connection()
        cur = conn.cursor()
        import_data_files(cur, dirname)

        QMessageBox.information(
            None, "Import effectué", "Import effectué avec succès")

    def sync_db(self):
        if not save_layers_edits():
            return
        from .sync import export_data_files, import_data_files
        from .ui.remote_connection import RemoteConnectionDlg
        import shutil
        import psycopg2

        while True:
            dlg = RemoteConnectionDlg()
            if dlg.exec_() != QDialog.Accepted:
                return

            try:
                remote_conn = psycopg2.connect(dlg.conn_string())
            except psycopg2.OperationalError:
                QMessageBox.critical(
                    None, "Problème de connexion",
                    "Impossible de se connecter à la base distante, \
vérifiez les paramètres de connexion")
            else:
                break
        dirname = tempfile.mkdtemp()
        remote_cur = remote_conn.cursor()

        local_conn = get_connection()
        local_cur = local_conn.cursor()

        QApplication.setOverrideCursor(Qt.WaitCursor)
        # export local data
        export_data_files(local_cur, dirname)

        # import in remote db
        import_data_files(remote_cur, dirname)

        # delete temp directory
        shutil.rmtree(dirname)
        QApplication.restoreOverrideCursor()

        QMessageBox.information(
            None, "Synchronisation effectuée", "Synchronisation effectuée")

    def new_ef(self, layer, point_clicked, parcel):
        f_to_valid = self.layers.get_feature(
            'gis', 'ef_parcels_to_valid', parcel['id'], id_key='parcel_id')
        curr_ef = self.layers.get_feature(
            'gis', 'current_ef_parcels', parcel['id'], id_key='parcel_id')

        if f_to_valid or curr_ef:  # the import of bm to ef/tf has already have done
            self.iface.messageBar().pushCritical(
                'Erreur',
                'Cette parcelle a déjà été importée!')
        else:
            conn = get_connection()
            cur = conn.cursor()

            cur.execute(
                "select gis.create_property_survey('{}');".format(
                    parcel['id']))
            cur.execute(
                "select gis.create_title_survey('{}');".format(parcel['id']))
            conn.commit()

            self.mapCanvas.refreshAllLayers()
