# -*- coding: utf-8 -*-

import os
from PyQt5 import uic
from PyQt5.QtCore import Qt, QAbstractTableModel, QDate, QVariant
from PyQt5.QtGui import QIcon, QGuiApplication
from PyQt5.QtWidgets import QDialog, QWidget, \
    QComboBox, QLineEdit, QHBoxLayout, QCheckBox, QToolButton, \
    QDateEdit, QMessageBox

from qgis.core import QgsFeatureRequest, QgsProject, QgsSettings
from .db import quote_nullable

import sip

SEARCH_BY_ATTRIBUTES_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'search_by_attribute.ui'))

# convert raw value to human-readable value
def display_value(value, type_name, ew, enumerations):
    if type_name == "bool":
        return "Oui" if value == "true" else "Non"
    elif ew.type() == "ValueRelation":
        enum = enumerations[type_name]
        vv = [v for v, k in enum if k == value]
        return vv[0] if vv else ""
    if isinstance(value, QDate):
        return value.toString("yyyy-MM-dd")
    if value is None or (isinstance(value, QVariant) and value.isNull()):
        return ""
    return value


class FilterClauseWidget(QWidget):
    def __init__(self, fields, allowed, enumerations, first=False, parent=None):
        self.enumerations = enumerations
        self.allowed = allowed
        
        QWidget.__init__(self, parent)
        self._and_or = QComboBox(self)
        self._field = QComboBox(self)
        self._operator = QComboBox(self)
        self._value_edit = QLineEdit(self)
        self._value_check = QCheckBox(self)
        self._value_combo = QComboBox(self)
        self._value_date = QDateEdit(self)
        self._value_date.setCalendarPopup(True)
        self._value_date.setDisplayFormat("dd/MM/yyyy")
        self._remove_btn = QToolButton(self)
        self._remove_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "delete.svg")))
        self._include_null_check = QCheckBox("Inclure valeurs nulles", self)
        self._layout = QHBoxLayout()
        self._layout.addWidget(self._remove_btn)
        self._layout.addWidget(self._and_or)
        self._layout.addWidget(self._field)
        self._layout.addWidget(self._operator)
        self._layout.addWidget(self._value_edit)
        self._layout.addWidget(self._value_combo)
        self._layout.addWidget(self._value_check)
        self._layout.addWidget(self._value_date)
        self._layout.addWidget(self._include_null_check)
        self._layout.addStretch()
        self.setLayout(self._layout)

        self._value = None

        self._and_or.addItem("ET")
        self._and_or.addItem("OU")
        self._operator.addItem("=")

        self._remove_btn.setVisible(not first)
        self._and_or.setVisible(not first)

        possible_field_types = list(self.enumerations.keys()) + ['float8', 'int4', 'bool', 'date', 'text']
        combo = []
        for idx, f in enumerate(fields):
            if f.typeName() not in possible_field_types:
                continue
            if f.name() not in allowed:
                continue
            n = f.alias() if f.alias() != "" else f.name()
            combo.append((n, idx, f.name()))
        combo = sorted(combo, key=lambda p: p[0])
        for alias, idx, name in combo:
            self._field.addItem(alias, (idx, name))
        self._fields = fields

        self._field.currentIndexChanged[int].connect(self.on_field_changed)
        self.on_field_changed()

    def and_or(self):
        return ("AND", "OR")[self._and_or.currentIndex()]

    def include_nulls(self):
        return self._include_null_check.isChecked()

    def field(self):
        # return field's db name
        return self._field.currentData()[1]

    def field_indexes(self):
        # return selected field indexes
        return [self._field.itemData(i)[0] for i in range(self._field.count())]

    def operator(self):
        return self._operator.currentData()

    def value(self):
        if isinstance(self._value, QLineEdit):
            t = self._fields[self._field.currentData()[0]].typeName()
            if t == "int4":
                try:
                    return int(self._value.text())
                except ValueError:
                    return 0
            elif t == "float8":
                try:
                    return float(self._value.text())
                except ValueError:
                    return 0.0
            return self._value.text()
        elif isinstance(self._value, QComboBox):
            return self._value.currentData()
        elif isinstance(self._value, QCheckBox):
            return "t" if self._value.isChecked() else "f"
        elif isinstance(self._value, QDateEdit):
            return self._value.date().toString(Qt.ISODate)

    def remove_btn(self):
        return self._remove_btn

    def on_field_changed(self):
        idx = self._field.currentData()[0]
        t = self._fields[idx].typeName()
        self._operator.clear()
        self._value_combo.clear()
        self._operator.addItem("égal(e) à", "=")
        self._operator.addItem("différent(e) de", "<>")
        if t == "text":
            self._value_edit.show()
            self._value_combo.hide()
            self._value_check.hide()
            self._value_date.hide()
            self._operator.addItem("LIKE", "like")
            self._operator.addItem("ILIKE", "ilike")
            self._value = self._value_edit
        elif t == "bool":
            self._value_edit.hide()
            self._value_combo.hide()
            self._value_check.show()
            self._value_date.hide()
            self._value = self._value_check
        elif t == "float8" or t == "int4":
            self._value_edit.show()
            self._value_combo.hide()
            self._value_check.hide()
            self._value_date.hide()
            self._operator.addItem("inférieur(e) à", "<")
            self._operator.addItem("inférieur(e) ou égal(e) à", "<=")
            self._operator.addItem("supérieur(e) à", ">")
            self._operator.addItem("supérieur(e) ou égal(e) à", ">=")
            self._value = self._value_edit
        elif t == "date":
            self._value_edit.hide()
            self._value_combo.hide()
            self._value_check.hide()
            self._value_date.show()
            self._value_date.setDate(QDate.currentDate())
            self._operator.addItem("antérieur(e) à", "<")
            self._operator.addItem("postérieur(e) à", ">")
            self._value = self._value_date
        elif t in self.enumerations:
            # We may have an enumerate type here, propose a combo with all the possible values
            self._value_edit.hide()
            self._value_combo.show()
            self._value_check.hide()
            self._value_date.hide()
            self._value = self._value_combo
            for user_value, db_value in self.enumerations[t]:
                self._value_combo.addItem(user_value, db_value)

class SearchByAttributeDialog(QDialog, SEARCH_BY_ATTRIBUTES_W):
    def __init__(self, layer, canvas, hidden_fields, display_whole_table=False, parent=None):
        super(SearchByAttributeDialog, self).__init__(parent)
        self.setupUi(self)
        self.add_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "add.svg")))
        self.search_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "search.svg")))
        self.attributes_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "attributes.svg")))

        self._layer = layer
        self._canvas = canvas

        # get values from Value Relation Widgets
        self.enumerations = {}
        self.allowed_fields = []
        for idx, f in enumerate(self._layer.fields()):
            ew = self._layer.editorWidgetSetup(idx)
            if ew.type() == "Hidden":
                continue
            if f.name() in hidden_fields:
                continue
            if ew.type() == "ValueRelation":
                l = QgsProject.instance().mapLayer(ew.config()['Layer'])
                en = [(v[ew.config()['Value']], v[ew.config()['Key']])for v in l.getFeatures()]
                self.enumerations[f.typeName()] = en
            # add to allowed fields
            self.allowed_fields.append(f.name())

        self.copy_btn.clicked.connect(self.on_copy)

        self.search_btn.clicked.connect(self.on_search)

        self.attributes_btn.clicked.connect(self.on_attributes)

        self._filters = []

        self.add_btn.clicked.connect(self.on_add_filter)
        self.on_add_filter(first=True)

        s = QgsSettings()
        self.zoom_on_selection.setChecked(s.value("PAGGF/zoom_on_selection", "false") == "true")
        def change_zoom():
            s.setValue("PAGGF/zoom_on_selection", self.zoom_on_selection.isChecked())
        self.zoom_on_selection.stateChanged.connect(change_zoom)

        # display the whole table on startup if needed
        if display_whole_table:
            self.on_attributes()

    def on_copy(self):
        try:
            rows = self.result_table.selectionModel().selectedRows()
        except AttributeError:
            QMessageBox.warning(self, "Table non initialisée.", "Veuillez d'abord effectuer une recherche ou afficher la table.")
            return

        # get feature ids from selection
        ids = [self.result_table.model().data(r.sibling(r.row(), 0), Qt.UserRole) for r in rows]
        if len(ids) > 0:
            clip = QGuiApplication.clipboard()
            s = ""
            # header
            fields = self._layer.fields()
            for fi in self._filters[0].field_indexes():
                s += "'"
                s += fields[fi].alias() if fields[fi].alias() else fields[fi].name()
                s += "'"
                s += '\t'
            s += '\n'

            req = QgsFeatureRequest()
            req.setFilterFids(ids)
            data = [f for f in self._layer.getFeatures(req)]
            for feature in data:
                attributes = feature.attributes()
                for fi in self._filters[0].field_indexes():
                    value = display_value(attributes[fi],
                                          self._layer.fields()[fi].typeName(),
                                          self._layer.editorWidgetSetup(fi),
                                          self.enumerations)
                    if type(value) not in (float,int):
                        s += "'" + str(value) + "'"
                    else:
                        s += str(value)
                    s += '\t'
                s += '\n'

            clip.setText(s)
        else:
            QMessageBox.warning(self, "Aucune ligne sélectionnée.", "Veuillez sélectionner au moins une ligne à copier.")

    def on_add_filter(self, first=False):
        clause_widget = FilterClauseWidget(self._layer.fields(), self.allowed_fields, self.enumerations, first, parent=self)
        def on_remove():
            self.on_remove_filter(clause_widget)
        clause_widget.remove_btn().clicked.connect(on_remove)

        self._filters.append(clause_widget)

        self.filter_layout.addWidget(clause_widget)

    def on_remove_filter(self, widget):
        self._filters.remove(widget)
        self.filter_layout.removeWidget(widget)
        widget.hide()
        sip.delete(widget)

    def on_search(self):
        where = ""
        first = True
        for f in self._filters:
            if not first:
                where += f.and_or() + " "
            where += '"{}" {} {} '.format(f.field(), f.operator(), quote_nullable(f.value()))
            if f.include_nulls():
                where += 'or "{}" is null '.format(f.field())
            first = False

        print("where", where, "layer", self._layer.name())
        req = QgsFeatureRequest().setFilterExpression(where)
        req.setSubsetOfAttributes(self._filters[0].field_indexes())
        self.result_table.setModel(TableModel(self._layer, req, self.enumerations))
        self.result_table.selectionModel().selectionChanged.connect(self.on_selection_changed)

    def on_attributes(self):
        req = QgsFeatureRequest()
        req.setSubsetOfAttributes(self._filters[0].field_indexes())
        self.result_table.setModel(TableModel(self._layer, req, self.enumerations))
        self.result_table.selectionModel().selectionChanged.connect(self.on_selection_changed)

    def on_selection_changed(self, selected, deselected):
        rows = self.result_table.selectionModel().selectedRows()
        # get feature ids from selection
        ids = [self.result_table.model().data(r.sibling(r.row(), 0), Qt.UserRole) for r in rows]
        self._layer.selectByIds(ids)
        if self.zoom_on_selection.isChecked():
            self._canvas.zoomToSelected(self._layer)

class TableModel(QAbstractTableModel):
    def __init__(self, layer, request, enumerations):
        QAbstractTableModel.__init__(self)
        self.page_size = 100
        self._layer = layer
        self.__data = [f for f in layer.getFeatures(request)]
        self._attributes = request.subsetOfAttributes()
        self._enumerations = enumerations

    def flags(self, idx):
        return Qt.NoItemFlags | Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def data(self, idx, role=Qt.DisplayRole):
        feat = self.__data[idx.row()]
        if idx.column() == 0 and role == Qt.UserRole:
            # id
            return feat.id()
        elif role == Qt.DisplayRole:
            idx = self._attributes[idx.column()]
            return display_value(feat.attributes()[idx],
                                 self._layer.fields()[idx].typeName(),
                                 self._layer.editorWidgetSetup(idx),
                                 self._enumerations)
        return None

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            f = self._layer.fields()[self._attributes[section]]
            return f.alias() if f.alias() != '' else f.name()
        return QAbstractTableModel.headerData(self, section, orientation, role)

    def rowCount(self, parent):
        return len(self.__data)

    def columnCount(self, parent):
        return len(self._attributes)
            
