import tempfile
import os.path
from io import BytesIO
from PyQt5 import uic
from PyQt5.QtWidgets import QDialog, QFileDialog
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtCore import QUrl
from ..b64_conv import B64Convertor
from ..deps.PyPDF2__v1_26_0.PyPDF2 import PdfFileMerger
from .. import paggf_version
from qgis.core import QgsSettings

FORM_CLASS, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ALL_merge_pv_and_signatures.ui'))


class ReportMergeDlg(QDialog, FORM_CLASS):
    def __init__(self, parcel, layer, version, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi(self)
        self.parcel = parcel
        self.layer = layer
        self.version = version
        self.signatures_filename = '<pas de signatures>'

        self.load_signatures_btn.clicked.connect(self.load_signatures_slot)
        self.merge_signature_and_report.clicked.connect(
            self.merge_signature_and_report_slot)
        self.download_signed_report.clicked.connect(
            self.download_signed_report_slot)
        self.see_signatures.clicked.connect(self.see_signatures_slot)

    def load_signatures_slot(self):
        s = QgsSettings("PAGGF", "PAGGF")
        last_signature_file_path = s.value("lastPAGGFSigantureFile", ".")

        if not os.path.exists(last_signature_file_path):
            last_signature_file_path = os.getcwd()

        self.signatures_filename, _ = QFileDialog.getOpenFileName(
            None, "Choisissez le fichier des signatures",
            last_signature_file_path, "Fichier pdf (*.pdf)")

        if self.signatures_filename == '':
            return

        s.setValue("lastPAGGFSigantureFile", last_signature_file_path)
        b64data = B64Convertor.base64_from_file(self.signatures_filename)

        # Save the generated pv in the database
        self.layer.startEditing()

        self.layer.changeAttributeValue(
            self.parcel.id(),
            self.layer.fields().indexFromName(self.signatures_field),
            b64data)

        self.parcel = self.layer.getFeature(self.parcel.id())
        self.update_display()

    def merge_signature_and_report_slot(self):
        self.download_signed_report.setEnabled(False)
        self.repaint()

        merger = PdfFileMerger()

        pv_fn, pv_file_content = B64Convertor.base64_decode(
            self.parcel["doc_generated_report"])
        pv_file_stream = BytesIO(pv_file_content)
        merger.append(pv_file_stream)

        signature_fn, signature_filecontent = B64Convertor.base64_decode(
            self.parcel[self.signatures_field])
        signature_stream = BytesIO(signature_filecontent)
        merger.append(signature_stream)

        if self.version == paggf_version.CID:
            pv_w_signatues_filename = os.path.join(
                tempfile.gettempdir(),
                '{}_{}_PVID_SIGNE.pdf'.format(
                    self.parcel['idc_parcel_code'],
                    self.parcel['hill_district_name']))
        else:  # EF
            pv_w_signatues_filename = os.path.join(
                tempfile.gettempdir(),
                '{}_{}_EF_SIGNE.pdf'.format(
                    self.parcel['idc_parcel_code'],
                    self.parcel['hill_district_name']))

        with open(pv_w_signatues_filename, 'wb') as pv_w_signatures_file:
            merger.write(pv_w_signatures_file)

        b64data = B64Convertor.base64_from_file(pv_w_signatues_filename)
        self.layer.startEditing()

        self.layer.changeAttributeValue(
            self.parcel.id(),
            self.layer.fields().indexFromName("doc_signed_report"),
            b64data)

        self.parcel = self.layer.getFeature(self.parcel.id())
        self.update_display()

    def see_signatures_slot(self):
        tmpfileurl = B64Convertor.file_from_base64(
            self.parcel[self.signatures_field])
        QDesktopServices.openUrl(
            QUrl.fromLocalFile(tmpfileurl))

    def download_signed_report_slot(self):
        tmpfileurl = B64Convertor.file_from_base64(
            self.parcel["doc_signed_report"])
        QDesktopServices.openUrl(
            QUrl.fromLocalFile(tmpfileurl))

    def update_display(self):
        if self.parcel[self.signatures_field]:
            self.merge_signature_and_report.setEnabled(True)
            self.see_signatures.setEnabled(True)
        else:
            self.merge_signature_and_report.setEnabled(False)
            self.see_signatures.setEnabled(False)

        if self.parcel["doc_signed_report"]:
            self.download_signed_report.setEnabled(True)
            self.report_w_signature_save.show()
            self.no_report_w_signature_save.hide()
        else:
            self.download_signed_report.setEnabled(False)
            self.report_w_signature_save.hide()
            self.no_report_w_signature_save.show()

        self.signature_file_path.setText(self.signatures_filename)

        if self.parcel["doc_generated_report"]:
            self.merge_signature_and_report.show()
            self.no_report_generated.hide()
        else:
            self.merge_signature_and_report.hide()
            self.no_report_generated.show()

        self.repaint()

    def exec_(self):
        if self.parcel[self.signatures_field]:
            self.signatures_filename, _ = B64Convertor.base64_decode(
                self.parcel[self.signatures_field])

        self.update_display()
        self.exec()


class CIDReportMergeDlg(ReportMergeDlg):
    def __init__(self, parcel, layer, parent=None):
        super(CIDReportMergeDlg, self).__init__(
            parcel, layer, paggf_version.CID, parent)
        self.signatures_field = "doc_idc_report_signatures"


class EFReportMergeDlg(ReportMergeDlg):
    def __init__(self, parcel, layer, parent=None):
        super(EFReportMergeDlg, self).__init__(
            parcel, layer, paggf_version.EF, parent)
        self.signatures_field = "doc_report_signatures"
